#ifndef ENGINE_H_
#define ENGINE_H_

#include "Choice.h"
#include "Who.h"
#include "Human.h"
#include "Computer.h"
#include "Display.h"

/**
 * The game Engine.
 */
class Engine
{
public:
    /**
     * Constructor.
     * @param max_rounds the maximum number of rounds per game.
     */
    Engine(Human * const h, Computer * const c, Display * const d)
        : human(h), computer(c), winner(Who::NONE), display(d),
          round(0), max_rounds(0),
          human_wins(0), computer_wins(0), ties(0),
          human_choice(Choice::NONE),
          computer_choice(Choice::NONE),
          predicted_choice(Choice::NONE) {}

    /**
     * Destructor.
     */
    virtual ~Engine() {}

    /**
     * Get the number of the current round.
     * @return the round number.
     */
    int get_round() const { return round; }

    /**
     * Get the human's choice.
     * @return the human's choice.
     */
    Choice get_human_choice() const { return human_choice; }

    /**
     * Set the human's choice.
     * @param choice the choice to set.
     */
    void set_human_choice(const Choice choice) { human_choice = choice; }

    /**
     * Get the computer's choice.
     * @return the computer's choice.
     */
    Choice get_computer_choice() const { return computer_choice; }

    /**
     * Get the number of human wins.
     * @return the number of human wins.
     */
    int get_human_wins() const { return human_wins; }

    /**
     * Get the number of computer wins.
     * @return the number of computer wins.
     */
    int get_computer_wins() const { return computer_wins; }

    /**
     * Get the number of ties.
     * @return the number of ties.
     */
    int get_ties() const { return ties; }

    /**
     * Start the game.
     * @param max_rounds the maximum number of rounds per game.
     */
    void start_game(const int max_rounds);

    /**
     * End the bame.
     */
    void end_game();

    /**
     * Start the next round.
     */
    void next_round();

    /**
     * Do the rounds of a game.
     */
    void do_rounds();

    /**
     * Do a round.
     */
    void do_round();

    /**
     * Determine the winner.
     * @param human_choice the human's choice.
     * @param computer_choice the computer's choice.
     * @return the winner.
     */
    Who determine_winner(const Choice human_choice,
                         const Choice computer_choice);

private:
    Human *human;        // human player
    Computer *computer;  // computer player
    Who winner;          // the winner (or tie)

    Display *display;    // game display

    int round;           // current round
    int max_rounds;      // max rounds per game
    int human_wins;      // number of human wins
    int computer_wins;   // number of computer wins
    int ties;            // number of ties

    Choice human_choice;      // human's current choice
    Choice computer_choice;   // computer's current choice
    Choice predicted_choice;  // predicted human's choice
};

#endif /* ENGINE_H_ */
