#include <string>
#include <iostream>
#include <fstream>
#include "SmartChoice.h"
#include "Choice.h"
#include "Engine.h"

using namespace std;

SmartChoice::SmartChoice()
    : history_length(HISTORY_LENGTH),
      filename(FREQUENCY_FILE_NAME)
{
    readFrequencies();
}

SmartChoice::~SmartChoice()
{
    writeFrequencies();
}

Choice SmartChoice::makeChoice()
{
    appendHistory(engine->getPredictedHumanChoice());
    updateFrequency();
    appendHistory(engine->getPrevComputerChoice());

    // Make the choice based on a prediction if possible,
    // else just make a random choice.
    Choice prediction = predictHumanChoice();
    switch (prediction)
    {
        case Choice::ROCK:     return Choice::PAPER;
        case Choice::PAPER:    return Choice::SCISSORS;
        case Choice::SCISSORS: return Choice::ROCK;
        default:               return randomChooser.makeChoice();
    }
}

void SmartChoice::appendHistory(const Choice choice)
{
    if (choice != Choice::NONE)
    {
        cout << "    Appended " << to_char(choice) << endl;
        history.push_back(to_char(choice));
        if (history.length() > history_length) history.erase(0, 1);
        cout << "    History " << history << endl;
   }
}

void SmartChoice::updateFrequency()
{
    int length = history.length();
    if (length < history_length) return;

    auto it = frequencies.find(history);

    if (it == frequencies.end())
    {
        frequencies.insert(pair<string, int>(history, 1));
        cout << "    Inserted " << history << " size=" << frequencies.size() << endl;
    }
    else
    {
        it->second++;
        cout << "    Updated " << history << " to " << it->second << endl;
    }
}

Choice SmartChoice::predictHumanChoice()
{
    int length = history.length();
    int maxFreq = 0;

    // Use the recorded frequencies to predict the human's choice.
    if (length >= history_length)
    {
        string most_recent = history.substr(1);
        string r_string = most_recent + 'R';
        string p_string = most_recent + 'P';
        string s_string = most_recent + 'S';

        int rFreq = getFrequency(r_string);
        int pFreq = getFrequency(p_string);
        int sFreq = getFrequency(s_string);

        if (rFreq > maxFreq)
        {
            maxFreq = rFreq;
            predictedHumanChoice = Choice::ROCK;
        }

        if (pFreq > maxFreq)
        {
            maxFreq = pFreq;
            predictedHumanChoice = Choice::PAPER;
        }

        if (sFreq > maxFreq)
        {
            maxFreq = sFreq;
            predictedHumanChoice = Choice::SCISSORS;
        }
    }

    return predictedHumanChoice;
}

int SmartChoice::getFrequency(const string history) const
{
    auto it = frequencies.find(history);
    int freq = it != frequencies.end() ? it->second : 0;

    cout << "    Checked: " << history << ":" << freq << endl;
    return freq;
}

void SmartChoice::readFrequencies()
{
    ifstream freq_in;
    freq_in.open(filename);

    if (freq_in.fail())
    {
        cout << "File " << filename << " does not exist." << endl;
    }
    else
    {
        string str;
        int n = 0, count;

        cout << "Reading file " << filename << ": ";
        do
        {
            freq_in >> str >> count;
            if (!freq_in.eof())
            {
                frequencies[str] = count;
                n++;
            }
        } while (!freq_in.eof());

        cout << n << " records." << endl;
        freq_in.close();
    }
}

void SmartChoice::writeFrequencies()
{
    ofstream freq_out;
    freq_out.open(filename);

    if (freq_out.fail())
    {
        cout << "Cannot write file " << filename << endl;
    }
    else
    {
        int n = 0;

        cout << "Writing frequency file "<< filename << ": ";
        for (auto it = frequencies.begin(); it != frequencies.end(); it++)
        {
            freq_out << it->first << " " << it->second << endl;
            n++;
        }

        cout << n << " records." << endl;
        freq_out.close();
    }

    cout << endl;
}
